/*
 *   (C) Copyright IBM Corp. 2006
 *   (C) Copyright ALT Linux Ltd. 2012
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: LUKS
 * File: evms2/engine/plugins/luks/luks.c
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include <uuid/uuid.h>

#include "luks.h"
#include "utils.h"

static plugin_record_t	luks_plugin_record;
plugin_record_t		*my_plugin_record = &luks_plugin_record;
engine_functions_t	*EngFncs = NULL;

static int luks_setup(engine_functions_t *functions)
{
	EngFncs = functions;
	LOG_ENTRY();
	LOG_EXIT_INT(0);
	return 0;
};

static int luks_can_delete(storage_object_t *object)
{
	int rc = EINVAL;

	LOG_ENTRY();

	if (utils_isa_luks(object) == TRUE) rc = 0;

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_can_expand(storage_object_t *object, u_int64_t expand_limit, list_anchor_t expansion_points)
{
	// TODO
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
};

static int luks_can_shrink(storage_object_t *object, u_int64_t shrink_limit, list_anchor_t shrink_points)
{
	// TODO
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
};

static int luks_discover(list_anchor_t input_objects, list_anchor_t output_objects, boolean final_call)
{
	list_anchor_t luks_object_list;
	list_element_t itr;
	storage_object_t *object, *luks;
	luks_metadata_t metadata;
	int luks_object_count = 0, rc;

	LOG_ENTRY();

	luks_object_list = EngFncs->allocate_list();
	if (!luks_object_list)
	{
		EngFncs->concatenate_lists(output_objects, input_objects);
		LOG_EXIT_INT(0);
		return 0;
	}

	LIST_FOR_EACH(input_objects, itr, object)
	{
		LOG_DEBUG("Examining object %s\n", object->name);

		if (object->plugin == my_plugin_record || object->data_type != DATA_TYPE)
		{
			EngFncs->insert_thing(output_objects, object, INSERT_AFTER, NULL);
			continue;
		}

		rc = utils_read_and_check_metadata(object, &metadata);
		if (rc)
		{
			LOG_DEBUG("LUKS metadata not found on object %s\n", object->name);
			EngFncs->insert_thing(output_objects, object, INSERT_AFTER, NULL);
			continue;
		}

		luks = utils_malloc_luks_object();
		if (luks == NULL)
		{
			LOG_DEBUG("cannot allocate memory\n");
			EngFncs->insert_thing(output_objects, object, INSERT_AFTER, NULL);
			continue;
		}

		utils_fill_pdata_from_metadata((luks_private_data_t*)luks->private_data, &metadata);

		snprintf(luks->name, EVMS_NAME_SIZE, "%s_luks", object->name);

		memcpy(&luks->geometry, &object->geometry, sizeof(geometry_t));
		luks->size = object->size - 4096;	// TODO: get from pdata/metadata

		EngFncs->insert_thing(object->parent_objects, luks, INSERT_BEFORE, NULL);
		EngFncs->insert_thing(luks->child_objects, object, INSERT_BEFORE, NULL);

		utils_get_devmap_info(luks);

		if (luks->flags & SOFLAG_ACTIVE) EngFncs->insert_thing(luks_object_list, luks, INSERT_BEFORE, NULL);
		else luks->flags |= SOFLAG_CORRUPT;
	}

	luks_object_count = EngFncs->list_count(luks_object_list);
	EngFncs->merge_lists(output_objects, luks_object_list, NULL, NULL);
	EngFncs->destroy_list(luks_object_list);

	LOG_EXIT_INT(luks_object_count);
	return luks_object_count;
};

static int luks_create(list_anchor_t input_objects, option_array_t *options, list_anchor_t output_objects)
{
	int rc = 0;
	storage_object_t *luks;
	luks_private_data_t *pdata = NULL;
	storage_object_t *child = EngFncs->first_thing(input_objects, NULL);

	LOG_ENTRY();

	REQUIRE(EngFncs->list_count(input_objects) == 1);
	REQUIRE(child);

	luks = utils_malloc_luks_object();

	if (luks == NULL)
	{
		rc = ENOMEM;
		LOG_EXIT_INT(rc);
		return rc;
	}

	pdata = (luks_private_data_t*)luks->private_data;
	utils_get_create_options(options, pdata);

	REQUIRE(strlen(pdata->cipher) > 0);
/*	Password always empty now
 *
	REQUIRE(strlen(pdata->password) > 0);
*/

	snprintf(luks->name, EVMS_NAME_SIZE, "%s_luks", child->name);
	uuid_generate_random((unsigned char*)&pdata->uuid);
	sprintf(luks->uuid, "%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X", pdata->uuid.time_low, pdata->uuid.time_mid, pdata->uuid.time_high, pdata->uuid.clock_seq_high, pdata->uuid.clock_seq_low, pdata->uuid.node[0], pdata->uuid.node[1], pdata->uuid.node[2], pdata->uuid.node[3], pdata->uuid.node[4], pdata->uuid.node[5]);

	rc = EngFncs->register_name(luks->name);
	rc += EngFncs->register_name(luks->uuid);
	if (rc)
	{
		rc = EINVAL;
		utils_free_luks_object(luks);
		LOG_EXIT_INT(rc);
		return rc;
	}

	memcpy(&luks->geometry, &child->geometry, sizeof(geometry_t));
	luks->size = child->size - 4096;

	EngFncs->insert_thing(child->parent_objects, luks, INSERT_BEFORE, NULL);
	EngFncs->insert_thing(luks->child_objects, child, INSERT_BEFORE, NULL);
	EngFncs->insert_thing(output_objects, luks, INSERT_BEFORE, NULL);

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_delete(storage_object_t *object, list_anchor_t child_objects)
{
	LOG_ENTRY();
	utils_delete_object(object, child_objects, TRUE);
	LOG_EXIT_INT(0);
	return 0;
};

static int luks_discard(list_anchor_t objects)
{
	storage_object_t *object;
	list_element_t le;

	LOG_ENTRY();

	LIST_FOR_EACH(objects, le, object)
	{
		utils_delete_object(object, NULL, FALSE);
	}

	LOG_EXIT_INT(0);
	return 0;
};

static int luks_expand(storage_object_t	*object, storage_object_t *expand_object, list_anchor_t input_objects, option_array_t *options)
{
	// TODO
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
};

static int luks_shrink(storage_object_t *object, storage_object_t *shrink_object, list_anchor_t input_objects, option_array_t *options)
{
	// TODO
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
};

static int luks_replace_child(storage_object_t *object, storage_object_t *child, storage_object_t *new_child)
{
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
};

static int luks_add_sectors_to_kill_list(storage_object_t *object, lsn_t lsn, sector_count_t count)
{
	int rc = EINVAL;
	storage_object_t *child;
	struct plugin_functions_s *fncs;

	LOG_ENTRY();

	REQUIRE(utils_isa_luks(object) == TRUE);
	REQUIRE(lsn + count <= object->size);
	REQUIRE(EngFncs->list_count(object->child_objects) == 1);

	child = EngFncs->first_thing(object->child_objects, NULL);

	fncs = (struct plugin_functions_s *)child->plugin->functions.plugin;
	rc = fncs->add_sectors_to_kill_list(child, lsn, count);

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_commit_changes(storage_object_t *object, uint commit_phase)
{
	int rc = 0;
	luks_private_data_t *pdata;

	LOG_ENTRY();

	LOG_DEBUG("object->name = %s, commit_phase = %d\n", object->name, commit_phase);

	REQUIRE(utils_isa_luks(object) == TRUE);

	pdata = (luks_private_data_t*)object->private_data;

	if (commit_phase == FIRST_METADATA_WRITE)
	{
/*	Password always empty now
 *
		if (pdata->password[0] == 0)
		{
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}
*/

		rc = utils_create_luks(object, commit_phase);
		if (rc != 0)
		{
			LOG_ERROR("error, failed to commit object\n");
			rc = ENOMSG;
		}
		else object->flags &= ~SOFLAG_DIRTY;
	}

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_can_activate(storage_object_t *object)
{
	LOG_ENTRY();
	REQUIRE(utils_isa_luks(object) == TRUE);
	LOG_EXIT_INT(0);
	return 0;
};

static int luks_activate(storage_object_t *object)
{
	int rc = 0;
	char password[EVMS_NAME_SIZE + 1];
	password[0] = 0;
	luks_private_data_t *pdata;

	LOG_ENTRY();

	REQUIRE(utils_isa_luks(object) == TRUE);

	pdata = (luks_private_data_t*)object->private_data;

/*	Password always empty now
 *
	if (pdata->password[0] == 0)
	{
		REQUEST_PASSWORD(password, NULL);
		if (password[0] == 0)
		{
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}
		strncpy(pdata->password, password, EVMS_NAME_SIZE);
	}
*/

	if (rc == 0) rc = utils_activate_luks(object);
	memset(pdata->password, 0, EVMS_NAME_SIZE + 1);
	if (rc == 0)
	{
		object->flags &= ~SOFLAG_NEEDS_ACTIVATE;
		object->flags &= ~SOFLAG_CORRUPT;
		// TODO: need rediscover
	}

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_can_deactivate(storage_object_t *object)
{
	LOG_ENTRY();
	REQUIRE(utils_isa_luks(object) == TRUE);
	LOG_EXIT_INT(0);
	return 0;
};

static int luks_deactivate(storage_object_t *object)
{
	int rc;

	LOG_ENTRY();

	REQUIRE(object != NULL);
	REQUIRE(object->plugin == my_plugin_record);

	rc = utils_deactivate_luks(object);
	if (rc==0) object->flags &= ~SOFLAG_NEEDS_DEACTIVATE;

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_get_option_count(task_context_t *task)
{
	int count;

	LOG_ENTRY();

	switch (task->action)
	{
		case EVMS_Task_Create:
			count = LUKS_CREATE_OPTION_COUNT;
			break;
		// TODO: EVMS_Task_Expand
		// TODO: EVMS_Task_Shrink
		default:
			count = 0;
			break;
	}

	LOG_EXIT_INT(count);
	return count;
};

static int luks_init_task(task_context_t *context)
{
	int rc = EINVAL;

	LOG_ENTRY();

	REQUIRE(context != NULL);

	switch (context->action)
	{
		case EVMS_Task_Create:
			context->min_selected_objects = 1;
			context->max_selected_objects = 1;

			rc = utils_allocate_create_option_descriptors(context);
			if (rc == 0) rc = utils_get_acceptable_create_objects(context);

			break;
		// TODO: EVMS_Task_Expand
		// TODO: EVMS_Task_Shrink
		default:
			LOG_ERROR("context->action is unknown or unsupported\n");
			break;
	}

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_set_option(task_context_t *context, u_int32_t index, value_t *value, u_int32_t *info)
{
	int rc = EINVAL;

	LOG_ENTRY();

	REQUIRE(context != NULL);

	switch (context->action)
	{
		case EVMS_Task_Create:
			rc = utils_set_create_option(context, index, value, info);
			break;
		default:
			rc = ENOSYS;
			break;
	}

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_set_objects(task_context_t *context, list_anchor_t declined_objects, task_effect_t *effect)
{
	int rc = 0;

	LOG_ENTRY();

	REQUIRE(context != NULL);
	REQUIRE(declined_objects != NULL);
	REQUIRE(effect != NULL);

	switch (context->action)
	{
		case EVMS_Task_Create:
			rc = utils_set_create_object(context, declined_objects, effect);
			break;
		// TODO: EVMS_Task_Expand
		// TODO: EVMS_Task_Shrink
		default:
			rc = EINVAL;
			LOG_ERROR("error, context->action is unknown or unsupported\n");
			break;
	}

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_get_info(storage_object_t *object, char *UNUSED(name),  extended_info_array_t **info)
{
	int rc = EINVAL;
	extended_info_array_t *Info;
	luks_private_data_t *pdata;

	LOG_ENTRY();

	REQUIRE(utils_isa_luks(object) == TRUE);
	REQUIRE(info != NULL);

	pdata = (luks_private_data_t*)object->private_data;

	Info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + (LUKS_INFO_COUNT * sizeof(extended_info_t)));
	if (Info)
	{
		Info->info[LUKS_INFO_NAME_INDEX].name			= EngFncs->engine_strdup("Name");
		Info->info[LUKS_INFO_NAME_INDEX].title			= EngFncs->engine_strdup(_("Name"));
		Info->info[LUKS_INFO_NAME_INDEX].desc			= EngFncs->engine_strdup(_("A mapping name luks object."));
		Info->info[LUKS_INFO_NAME_INDEX].type			= EVMS_Type_String;
		Info->info[LUKS_INFO_NAME_INDEX].unit			= EVMS_Unit_None;
		Info->info[LUKS_INFO_NAME_INDEX].value.s		= EngFncs->engine_strdup(object->name);
		Info->info[LUKS_INFO_NAME_INDEX].collection_type	= EVMS_Collection_None;
		memset(&Info->info[LUKS_INFO_NAME_INDEX].group, 0, sizeof(group_info_t));
		Info->count += 1;

		Info->info[LUKS_INFO_SIZE_INDEX].name			= EngFncs->engine_strdup("Size");
		Info->info[LUKS_INFO_SIZE_INDEX].title			= EngFncs->engine_strdup(_("Size"));
		Info->info[LUKS_INFO_SIZE_INDEX].desc			= EngFncs->engine_strdup(_("The size of the object."));
		Info->info[LUKS_INFO_SIZE_INDEX].type			= EVMS_Type_Unsigned_Int64;
		Info->info[LUKS_INFO_SIZE_INDEX].unit			= EVMS_Unit_Sectors;
		Info->info[LUKS_INFO_SIZE_INDEX].value.ui64		= object->size;
		Info->info[LUKS_INFO_SIZE_INDEX].collection_type	= EVMS_Collection_None;
		memset(&Info->info[LUKS_INFO_SIZE_INDEX].group, 0, sizeof(group_info_t));
		Info->info[LUKS_INFO_SIZE_INDEX].flags			|= EVMS_EINFO_FLAGS_NO_UNIT_CONVERSION;
		Info->count += 1;

		Info->info[LUKS_INFO_CIPHER_INDEX].name			= EngFncs->engine_strdup("Cipher");
		Info->info[LUKS_INFO_CIPHER_INDEX].title		= EngFncs->engine_strdup(_("Cipher"));
		Info->info[LUKS_INFO_CIPHER_INDEX].desc			= EngFncs->engine_strdup(_("Cipher specification string."));
		Info->info[LUKS_INFO_CIPHER_INDEX].type			= EVMS_Type_String;
		Info->info[LUKS_INFO_CIPHER_INDEX].unit			= EVMS_Unit_None;
		Info->info[LUKS_INFO_CIPHER_INDEX].value.s		= EngFncs->engine_strdup(pdata->cipher);
		Info->info[LUKS_INFO_CIPHER_INDEX].collection_type	= EVMS_Collection_None;
		memset(&Info->info[LUKS_INFO_CIPHER_INDEX].group, 0, sizeof(group_info_t));
		Info->count += 1;

		*info = Info;
		rc = 0;
	}
	else
	{
		rc = ENOMEM;
		LOG_ERROR("unable to malloc memory for extended info array\n");
	}

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_get_plugin_info(char *UNUSED(descriptor_name), extended_info_array_t **info)
{
	int			rc = EINVAL;
	extended_info_array_t	*Info;
	char			version_string[64];
	char			required_engine_api_version_string[64];
	char			required_plugin_api_version_string[64];

	LOG_ENTRY();

	REQUIRE(info != NULL);

	*info = NULL;

	Info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + (LUKS_PLUGIN_INFO_COUNT * sizeof(extended_info_t)));
	if (Info)
	{
		Info->count = LUKS_PLUGIN_INFO_COUNT;
		sprintf(version_string, "%d.%d.%d", MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);
		sprintf(required_engine_api_version_string, "%d.%d.%d", my_plugin_record->required_engine_api_version.major, my_plugin_record->required_engine_api_version.minor, my_plugin_record->required_engine_api_version.patchlevel);
		sprintf(required_plugin_api_version_string, "%d.%d.%d", my_plugin_record->required_plugin_api_version.plugin.major, my_plugin_record->required_plugin_api_version.plugin.minor, my_plugin_record->required_plugin_api_version.plugin.patchlevel);

		Info->info[LUKS_PLUGIN_INFO_SNAME_INDEX].name			= EngFncs->engine_strdup("Short Name");
		Info->info[LUKS_PLUGIN_INFO_SNAME_INDEX].title			= EngFncs->engine_strdup(_("Short Name"));
		Info->info[LUKS_PLUGIN_INFO_SNAME_INDEX].desc			= EngFncs->engine_strdup(_("A short name given to this plug-in"));
		Info->info[LUKS_PLUGIN_INFO_SNAME_INDEX].type			= EVMS_Type_String;
		Info->info[LUKS_PLUGIN_INFO_SNAME_INDEX].unit			= EVMS_Unit_None;
		Info->info[LUKS_PLUGIN_INFO_SNAME_INDEX].value.s		= EngFncs->engine_strdup(my_plugin_record->short_name);
		Info->info[LUKS_PLUGIN_INFO_SNAME_INDEX].collection_type	= EVMS_Collection_None;
		memset(&Info->info[LUKS_PLUGIN_INFO_SNAME_INDEX].group, 0, sizeof(group_info_t));

		Info->info[LUKS_PLUGIN_INFO_LNAME_INDEX].name			= EngFncs->engine_strdup("Long Name");
		Info->info[LUKS_PLUGIN_INFO_LNAME_INDEX].title			= EngFncs->engine_strdup(_("Long Name"));
		Info->info[LUKS_PLUGIN_INFO_LNAME_INDEX].desc			= EngFncs->engine_strdup(_("A longer, more descriptive name for this plug-in"));
		Info->info[LUKS_PLUGIN_INFO_LNAME_INDEX].type			= EVMS_Type_String;
		Info->info[LUKS_PLUGIN_INFO_LNAME_INDEX].unit			= EVMS_Unit_None;
		Info->info[LUKS_PLUGIN_INFO_LNAME_INDEX].value.s		= EngFncs->engine_strdup(my_plugin_record->long_name);
		Info->info[LUKS_PLUGIN_INFO_LNAME_INDEX].collection_type	= EVMS_Collection_None;
		memset(&Info->info[LUKS_PLUGIN_INFO_LNAME_INDEX].group, 0, sizeof(group_info_t));

		Info->info[LUKS_PLUGIN_INFO_TYPE_INDEX].name			= EngFncs->engine_strdup("Type");
		Info->info[LUKS_PLUGIN_INFO_TYPE_INDEX].title			= EngFncs->engine_strdup(_("Plug-in Type"));
		Info->info[LUKS_PLUGIN_INFO_TYPE_INDEX].desc			= EngFncs->engine_strdup(_("There are various types of plug-ins, each responsible for some kind of storage object or logical volume."));
		Info->info[LUKS_PLUGIN_INFO_TYPE_INDEX].type			= EVMS_Type_String;
		Info->info[LUKS_PLUGIN_INFO_TYPE_INDEX].unit			= EVMS_Unit_None;
		Info->info[LUKS_PLUGIN_INFO_TYPE_INDEX].value.s			= EngFncs->engine_strdup(_("EVMS Feature"));
		Info->info[LUKS_PLUGIN_INFO_TYPE_INDEX].collection_type		= EVMS_Collection_None;
		memset(&Info->info[LUKS_PLUGIN_INFO_TYPE_INDEX].group, 0, sizeof(group_info_t));

		Info->info[LUKS_PLUGIN_INFO_VERSION_INDEX].name			= EngFncs->engine_strdup("Version");
		Info->info[LUKS_PLUGIN_INFO_VERSION_INDEX].title		= EngFncs->engine_strdup(_("Plug-in Version"));
		Info->info[LUKS_PLUGIN_INFO_VERSION_INDEX].desc			= EngFncs->engine_strdup(_("This is the version number of the plug-in."));
		Info->info[LUKS_PLUGIN_INFO_VERSION_INDEX].type			= EVMS_Type_String;
		Info->info[LUKS_PLUGIN_INFO_VERSION_INDEX].unit			= EVMS_Unit_None;
		Info->info[LUKS_PLUGIN_INFO_VERSION_INDEX].value.s		= EngFncs->engine_strdup(version_string);
		Info->info[LUKS_PLUGIN_INFO_VERSION_INDEX].collection_type	= EVMS_Collection_None;
		memset(&Info->info[LUKS_PLUGIN_INFO_VERSION_INDEX].group, 0, sizeof(group_info_t));

		Info->info[LUKS_PLUGIN_INFO_REQENGSVCVERSION_INDEX].name		= EngFncs->engine_strdup("Required Engine Services Version");
		Info->info[LUKS_PLUGIN_INFO_REQENGSVCVERSION_INDEX].title		= EngFncs->engine_strdup(_("Required Engine Services Version"));
		Info->info[LUKS_PLUGIN_INFO_REQENGSVCVERSION_INDEX].desc		= EngFncs->engine_strdup(_("This is the version of the Engine services that this plug-in requires. It will not run on older versions of the Engine services."));
		Info->info[LUKS_PLUGIN_INFO_REQENGSVCVERSION_INDEX].type		= EVMS_Type_String;
		Info->info[LUKS_PLUGIN_INFO_REQENGSVCVERSION_INDEX].unit		= EVMS_Unit_None;
		Info->info[LUKS_PLUGIN_INFO_REQENGSVCVERSION_INDEX].value.s		= EngFncs->engine_strdup(required_engine_api_version_string);
		Info->info[LUKS_PLUGIN_INFO_REQENGSVCVERSION_INDEX].collection_type	= EVMS_Collection_None;
		memset(&Info->info[LUKS_PLUGIN_INFO_REQENGSVCVERSION_INDEX].group, 0, sizeof(group_info_t));

		Info->info[LUKS_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].name		= EngFncs->engine_strdup("Required Engine Plug-in API Version");
		Info->info[LUKS_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].title		= EngFncs->engine_strdup(_("Required Engine Plug-in API Version"));
		Info->info[LUKS_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].desc		= EngFncs->engine_strdup(_("This is the version of the Engine plug-in API that this plug-in requires. It will not run on older versions of the Engine plug-in API."));
		Info->info[LUKS_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].type		= EVMS_Type_String;
		Info->info[LUKS_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].unit		= EVMS_Unit_None;
		Info->info[LUKS_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].value.s		= EngFncs->engine_strdup(required_plugin_api_version_string);
		Info->info[LUKS_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].collection_type	= EVMS_Collection_None;
		memset(&Info->info[LUKS_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].group, 0, sizeof(group_info_t));

		*info = Info;
		rc=0;
	}

	LOG_EXIT_INT(rc);
	return rc;
};

static int luks_read(storage_object_t *object, lsn_t lsn, sector_count_t count, void *buffer)
{
	LOG_ENTRY();
	LOG_EXIT_INT(0);
	return 0;
};

static int luks_write(storage_object_t *object, lsn_t lsn, sector_count_t count, void *buffer)
{
	LOG_ENTRY();
	LOG_EXIT_INT(0);
	return 0;
};

static int luks_get_plugin_functions(storage_object_t *object, function_info_array_t **functions)
{
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
};

static int luks_plugin_function(storage_object_t *object, task_action_t action, list_anchor_t objects, option_array_t *options)
{
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
};

static int luks_backup_metadata(storage_object_t *object)
{
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
};

static void luks_cleanup(void)
{
	int rc;
	storage_object_t *luks;
	luks_private_data_t *pdata;
	list_anchor_t object_list = NULL;
	list_element_t iter;

	LOG_ENTRY();

	rc = utils_get_luks_list(&object_list);
	if (rc==0)
	{
		LIST_FOR_EACH(object_list, iter, luks)
		{
			pdata = (luks_private_data_t *) luks->private_data;
			if (pdata)
			{
				free(pdata);
				luks->private_data = NULL;
			}
		}
		EngFncs->destroy_list(object_list);
	}

	LOG_EXIT_VOID();
};

static plugin_functions_t luks_functions =
{
	setup_evms_plugin		: luks_setup,
	cleanup_evms_plugin		: luks_cleanup,
	can_delete			: luks_can_delete,
	can_expand			: luks_can_expand,
	can_shrink			: luks_can_shrink,
	discover			: luks_discover,
	create				: luks_create,
	delete				: luks_delete,
	discard				: luks_discard,
	expand				: luks_expand,
	shrink				: luks_shrink,
	replace_child			: luks_replace_child,
	add_sectors_to_kill_list	: luks_add_sectors_to_kill_list,
	commit_changes			: luks_commit_changes,
	activate			: luks_activate,
	can_activate			: luks_can_activate,
	can_deactivate			: luks_can_deactivate,
	deactivate			: luks_deactivate,
	get_option_count		: luks_get_option_count,
	init_task			: luks_init_task,
	set_option			: luks_set_option,
	set_objects			: luks_set_objects,
	get_info			: luks_get_info,
	get_plugin_info			: luks_get_plugin_info,
	read				: luks_read,
	write				: luks_write,
	get_plugin_functions		: luks_get_plugin_functions,
	plugin_function			: luks_plugin_function,
	backup_metadata                 : luks_backup_metadata
};

static plugin_record_t luks_plugin_record =
{
	id:				EVMS_LUKS_PLUGIN_ID,
	version:			{
						major:		MAJOR_VERSION,
						minor:		MINOR_VERSION,
						patchlevel:	PATCH_LEVEL
					},
	required_engine_api_version:	{
						major:		15,
						minor:		0,
						patchlevel:	0
					},
	required_plugin_api_version:	{
						plugin:	{
								major:		13,
								minor:		1,
								patchlevel:	0
							}
					},
	short_name:			EVMS_LUKS_PLUGIN_SHORT_NAME,
	long_name:			EVMS_LUKS_PLUGIN_LONG_NAME,
	oem_name:			EVMS_ALT_LINUX_OEM_NAME,
	functions:			{
						plugin:	&luks_functions
					},
	container_functions:		NULL
};

plugin_record_t *evms_plugin_records[] =
{
	&luks_plugin_record,
	NULL
};
