/*
 *   (C) Copyright IBM Corp. 2001, 2005
 *   (C) Copyright ALT Linux Ltd. 2011
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: mdregmgr
 * File: raid10_mgr.c
 *
 * Description: This file contains all of the required engine-plugin APIs
 *              for the Raid10 MD region manager.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include <malloc.h>

#include "md.h"
#include "raid10_mgr.h"

#define my_plugin_record raid10_plugin

static void raid10_free_private_data(md_volume_t * volume)
{
	raid10_conf_t * conf = (raid10_conf_t *)volume->private_data;
	int i;
	struct strip_zone *zone;

	LOG_ENTRY();

	if (!conf) {
		LOG_WARNING("Nothing to free!!!.\n");
		LOG_EXIT_VOID();
		return;
	}

	if (conf->strip_zone) {
		for (i=0; i<conf->nr_strip_zones; i++) {
			zone = conf->strip_zone + i;
			if (zone->dev) {
				EngFncs->engine_free(zone->dev);
				zone->dev = NULL;
			}
		}
		EngFncs->engine_free(conf->strip_zone);
		conf->strip_zone = NULL;
	}
	if (conf->hash_table) {
		EngFncs->engine_free(conf->hash_table);
		conf->hash_table = NULL;
	}
	EngFncs->engine_free(volume->private_data);
	volume->private_data = NULL;
	LOG_EXIT_VOID();
}

static int raid10_setup_evms_plugin(engine_functions_t * functions)
{
	int rc = 0;

	if (!functions) return EINVAL;

	EngFncs = functions;

	my_plugin = raid10_plugin;
	LOG_ENTRY();
	rc = md_register_name_space();

	if (rc != 0) LOG_SERIOUS("Failed to register the MD name space.\n");

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_can_delete(storage_object_t * UNUSED(region))
{
	LOG_ENTRY();
	LOG_EXIT_INT(0);
	return 0;
}

static int raid10_can_expand(storage_object_t * region, u_int64_t expand_limit, list_anchor_t expansion_points)
{
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	LOG_EXIT_INT(EINVAL);
	return EINVAL;
}

static int raid10_can_shrink(storage_object_t * region, u_int64_t shrink_limit, list_anchor_t shrink_points)
{
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	LOG_EXIT_INT(EINVAL);
	return EINVAL;
}

static int raid10_can_replace_child(storage_object_t *region, storage_object_t *child, storage_object_t *new_child)
{
	int rc;
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	rc = md_can_replace_child(region, child, new_child);
	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_init_region(md_volume_t *vol, storage_object_t *region, boolean final_call)
{
	int rc = 0;
	md_member_t *member;
	list_element_t iter;
	mdu_array_info_t info;

	LOG_ENTRY();

	LIST_FOR_EACH(vol->members, iter, member) {
		if (member->obj) {
			md_append_region_to_object(region, member->obj);
		} else {
			LOG_MD_BUG();
		}
	}

	region->size = md_volume_calc_size(vol);
	region->data_type = DATA_TYPE;
	region->plugin = raid10_plugin;
	region->private_data = (void *)vol;
	region->dev_major = MD_MAJOR;
	region->dev_minor = vol->md_minor;

	vol->flags |= MD_DISCOVERED;
	vol->region = region;

	md_get_kernel_info(region, &info);
	if (final_call) {
		if (region->flags & SOFLAG_ACTIVE) {
			rc = md_analyze_active_region(vol);
		} else {
			md_analyze_volume(vol);
			md_fix_dev_major_minor(vol, TRUE);
		}
	} else {
		// Needs to validate later
		vol->flags |= MD_NEEDS_VALIDATE;
	}

	if (vol->flags & MD_CORRUPT) {
		region->flags |= SOFLAG_CORRUPT;
	}

	LOG_DETAILS("Region [%s] has been created (%s, %s, %s)\n",
		    region->name,
		    (vol->flags & MD_DISCOVERED) ? "discovered" : "BUG: not discovered",
		    (region->flags & SOFLAG_ACTIVE) ? "active" : "inactive",
		    (vol->flags & MD_DEGRADED) ? "degraded" : ((vol->flags & MD_CORRUPT) ? "corrupt" : "normal"));

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_create_region(md_volume_t *vol, list_anchor_t output_list, boolean final_call)
{
	int rc = 0;
	int length;
	storage_object_t * region;

	LOG_ENTRY();

	if (!vol->sb) {
		LOG_MD_BUG();
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	if (!vol->active_disks) {
		if (final_call) {
			vol->flags |= MD_CORRUPT;
			length = sprintf(message_buffer,
					 _("RAID10 region %s is corrupt. "
					   "  The number of raid disks for a full functional array is %d."
					   "  The number of active disks is %d."),
					 vol->name, vol->raid_disks, vol->active_disks);
			md_queue_corrupt_message(vol, message_buffer, length);
		} else {
			LOG_DEBUG("Volume %s does not have any active disks, delaying discovery.\n",
				  vol->name);
			LOG_EXIT_INT(0);
			return 0;
		}
	}

	rc = EngFncs->allocate_region(vol->name, &region);
	if (rc) {
		LOG_WARNING("Region %s is already created, try new name.\n", vol->name);
		rc = md_volume_get_alternative_name(vol, 255);
		if (!rc) {
			LOG_WARNING("Trying tnew region name: %s...\n", vol->name);
			rc = EngFncs->allocate_region(vol->name, &region);
			if (!rc) {
				LOG_WARNING("OK. got it.\n");
			} else {
				LOG_ERROR("Give up.\n");
			}
		} else {
			LOG_CRITICAL("The MD region name space has been exausted.\n");
			LOG_EXIT_INT(rc);
			return rc;
		}
	}

	if (!rc) {
		rc = raid10_init_region(vol, region, final_call);
		md_add_object_to_list(region, output_list);
	}

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_rediscover_region(storage_object_t * region, boolean final_call)
{
	int rc = 0;
	md_volume_t   *vol;
	list_anchor_t children;
	list_anchor_t output_list;
	md_member_t *member;
	list_element_t iter;
	storage_object_t *obj;
	u_int32_t md_minor;

	LOG_ENTRY();

	if (region == NULL) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	vol = region->private_data;

	if (md_is_recovery_running(region)) {
		/*
		 * If array is syncing, don't rediscover
		 * Set flag to rediscover the array when it's done.
		 */
		LOG_DEBUG("MD array %s is syncing, skipping rediscovery.\n", vol->name);
		vol->flags |= MD_ARRAY_SYNCING;
		LOG_EXIT_INT(0);
		return 0;
	}

	LOG_DEBUG("About to rediscover volume %s.\n", vol->name);

	md_minor = vol->md_minor; // Save the minor for later comparison

	children = EngFncs->allocate_list();
	output_list = EngFncs->allocate_list();
	md_clear_child_list(region, children);

	LIST_FOR_EACH(vol->members, iter, member) {
		if (member->obj) {
			EngFncs->insert_thing(children, member->obj, INSERT_AFTER | EXCLUSIVE_INSERT, NULL);
		}
	}
	md_free_volume(vol);
	LIST_FOR_EACH(children, iter, obj) {
		LOG_DEBUG("   Rediscover on this object: %s.\n", obj->name);
	}
	md_discover_volumes(children, output_list);
	region->private_data = NULL;

	vol = volume_list_head;
	while (vol != NULL) {
		if ((!(vol->flags & MD_DISCOVERED)) && (vol->personality == RAID10) && (vol->md_minor == md_minor)) {
			region->flags &= ~(SOFLAG_DIRTY | SOFLAG_CORRUPT | SOFLAG_NEEDS_ACTIVATE | SOFLAG_NEEDS_DEACTIVATE | SOFLAG_ACTIVE);
			rc = raid10_init_region(vol, region, final_call);
		}
		vol = vol->next;
	}

	EngFncs->destroy_list(children);
	EngFncs->destroy_list(output_list);

	if (region->private_data == NULL) {
		LOG_MD_BUG();
		EngFncs->free_region(region);
		rc = ENODEV;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

int raid10_discover_regions(list_anchor_t output_list, int *count, boolean final_call)
{
	int rc = 0;
	md_volume_t * volume = volume_list_head;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

retry:
	while (volume != NULL) {
		if ((!(volume->flags & MD_DISCOVERED)) && (volume->personality == RAID10)) {
		       	rc = raid10_create_region(volume, output_list, final_call);
		       	if (volume->flags & MD_DISCOVERED) {
		       		*count = *count + 1;
		       	}
		}
		if ((volume->flags & MD_DISCOVERED) && (volume->flags & MD_NEEDS_REDISCOVER) && (volume->personality == RAID10)) {
			volume->flags &= ~MD_NEEDS_REDISCOVER;
			rc = raid10_rediscover_region(volume->region, final_call);
			if (!(volume->flags & MD_NEEDS_REDISCOVER)) {
				/* The region was rediscovered,
				 * go back to check everything again.
				 */
				goto retry;
			}
		}
		volume = volume->next;
	}
	if (final_call) {
		md_display_corrupt_messages(RAID10);
	}

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_discover(list_anchor_t input_list, list_anchor_t output_list, boolean final_call)
{
	int count = 0;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	// Parameter check
	if (!input_list || !output_list) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	if (final_call) {
		md_discover_final_call(input_list, output_list, &count);
	} else {
		md_discover_volumes(input_list, output_list);
		LOG_DETAILS("PV discovery complete.\n");

		// LV discovery and exporting
		raid10_discover_regions(output_list, &count, final_call);
		LOG_DETAILS("RAID10 volume discovery complete.\n");
	}
	LOG_EXIT_INT(count);
	return count;
}

static int raid10_get_create_options(option_array_t *options, u_int32_t *chunksize, md_sb_ver_t *sb_ver, u_int32_t *layout)
{
	uint i;
	int rc = 0;
	boolean ver1_superblock = FALSE;
	int layout_near = RAID10_DEFAULT_LAYOUT_NEAR;

	LOG_ENTRY();

	for (i = 0; i < options->count; i++) {

		if (options->option[i].is_number_based) {

			switch (options->option[i].number) {
			case RAID10_CREATE_OPT_SB1_INDEX:
				ver1_superblock = options->option[i].value.b;
				break;

			case RAID10_CREATE_OPT_CHUNK_SIZE_INDEX:
				// Option is in 1K byte blocks
				*chunksize = 2 * options->option[i].value.ui32;
				break;

			case RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX:
				layout_near = options->option[i].value.ui8;
				break;

			default:
				break;
			}

		} else {

			if (strcmp(options->option[i].name, RAID10_CREATE_OPT_CHUNK_SIZE_NAME) == 0) *chunksize = 2 * options->option[i].value.ui32;
			else if (strcmp(options->option[i].name, RAID10_CREATE_OPT_SB1_NAME) == 0) ver1_superblock =  options->option[i].value.b;
			else if (strcmp(options->option[i].name, RAID10_CREATE_OPT_LAYOUT_NEAR_NAME) == 0) layout_near = options->option[i].value.ui8;
		}
	}

	if (ver1_superblock == TRUE) {
		sb_ver->major_version = MD_SB_VER_1;
		sb_ver->minor_version = 0;
		sb_ver->patchlevel = 0;
	} else {
		sb_ver->major_version = MD_SB_VER_0;
		sb_ver->minor_version = 90;
		sb_ver->patchlevel = 0;
	}

	*layout=layout_near+(1<<8);

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_create_new_region(md_volume_t * vol, list_anchor_t output_list)
{
	int rc=0;
	md_member_t *member;
	list_element_t iter;
	storage_object_t *region;

	LOG_ENTRY();
	rc = EngFncs->allocate_region(vol->name, &region);
	if (rc) {
		LOG_ERROR("Region %s is already created (rc=%d).\n",
			  vol->name, rc);
	}

	if (!rc) {
		LOG_DEBUG("Creating new region %s: nr_disks=%d, raid_disks=%d,"
			  " spares=%d, actives=%d, working=%d\n",
			  vol->name, vol->nr_disks, vol->raid_disks,
			  vol->spare_disks, vol->active_disks, vol->working_disks);

		LIST_FOR_EACH(vol->members, iter, member) {
			md_append_region_to_object(region, member->obj);
		}
		region->size = md_volume_calc_size(vol);
		region->data_type = DATA_TYPE;
		region->plugin = raid10_plugin;
		region->private_data = (void *)vol;
		region->dev_major = MD_MAJOR;
		region->dev_minor = vol->md_minor;
		vol->region = region;
		region->flags |= SOFLAG_DIRTY;
		md_add_object_to_list(region, output_list);
	}
	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_create(list_anchor_t objects, option_array_t *options, list_anchor_t new_region_list)
{
	md_volume_t * volume = NULL;
	storage_object_t * object;
	unsigned long size = -1;
	list_element_t iter1, iter2;
	md_sb_ver_t sb_ver = {MD_SB_VER_0, 90, 0};
	md_member_t *member;
	u_int32_t chunksize = MD_DEFAULT_CHUNK_SIZE;
	u_int32_t layout = 0;
	int count;
	int rc = 0;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	// Parameter check
	if (!objects || !options || !new_region_list) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	count = EngFncs->list_count(objects);

	raid10_get_create_options(options, &chunksize, &sb_ver, &layout);

	if (count < (layout & 0xFF)) {
		MESSAGE(_("Must have at least %d object.\n"), (layout & 0xFF));
		rc = EINVAL;
		goto out;
	}

	if (count > MAX_SB_DISKS(&sb_ver)) {
		MESSAGE(_("Too many devices (%d) were specified."
		          "  For MD superblock version %d.%d, the maximum is %d.\n"),
		        count, sb_ver.major_version, sb_ver.minor_version,
		        MAX_SB_DISKS(&sb_ver));
		return EINVAL;
	}

	if (!(volume = md_allocate_volume())) {
		rc = ENOMEM;
		goto out;
	}

	rc = md_volume_get_available_name(volume, 256);
	if (rc) {
		goto error_free;
	}

	LIST_FOR_EACH(objects, iter1, object) {
		size = min(size, md_object_usable_size(object, &sb_ver, chunksize));
	}

	rc = md_init_sb(volume, &sb_ver, MD_LEVEL_RAID10, layout, size, chunksize);
	if (rc) {
		goto error_free;
	}
	// Add raid members
	LIST_FOR_EACH_SAFE(objects, iter1, iter2, object) {
		member = md_allocate_member(object);
		if (member) {
			// This will add the member and update the MD superblock.
			member->data_size = md_object_usable_size(object, &sb_ver, chunksize);
			member->flags |= (MD_MEMBER_NEW | MD_MEMBER_DISK_ACTIVE | MD_MEMBER_DISK_SYNC);
			rc = md_volume_add_new_member(volume, member);
			if (rc) {
				md_free_member(member);
				goto error_free;
			}
		} else {
			rc = ENOMEM;
		}
		if (rc) {
			goto error_free;
		}
		EngFncs->delete_element(iter1);
	}

	rc = raid10_create_new_region(volume, new_region_list);
	if (!rc) {
		volume->flags |= MD_DIRTY;
	}
error_free:
	if (rc && volume) {
		md_free_volume(volume);
	}
out:
	LOG_EXIT_INT(rc);
	return rc;
}

static int w_delete(storage_object_t *region, list_anchor_t children, boolean tear_down)
{
	md_volume_t * volume;
	int rc = 0;

	LOG_ENTRY();

	// Check that this region can be removed.
	if ((rc = raid10_can_delete(region))) {
		LOG_EXIT_INT(rc);
		return rc;
	}
	volume = region->private_data;

	// Remove the parent/child associations with the PVs
	md_clear_child_list(region, children);

	//Delete raid10 private_data
	raid10_free_private_data(volume);

	// Delete the volume.
	md_delete_volume(volume, tear_down);
	EngFncs->free_region(region);

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_delete(storage_object_t * region, list_anchor_t children)
{
	int           rc;

	my_plugin = raid10_plugin;
	LOG_ENTRY();
	rc = w_delete(region, children, TRUE);
	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_discard(list_anchor_t regions)
{
	storage_object_t * region;
	list_element_t le;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	LIST_FOR_EACH(regions, le, region) {
		w_delete(region, NULL, FALSE);
	}

	LOG_EXIT_INT(0);
	return 0;
}

static int raid10_expand(storage_object_t *region, storage_object_t *UNUSED(expand_object), list_anchor_t input_objects, option_array_t *UNUSED(options))
{
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	LOG_EXIT_INT(EINVAL);
	return EINVAL;
}

static int raid10_shrink(storage_object_t *region, storage_object_t *UNUSED(shrink_object), list_anchor_t input_objects, option_array_t *UNUSED(options))
{
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	LOG_EXIT_INT(EINVAL);
	return EINVAL;
}

static int raid10_replace_child(storage_object_t *region, storage_object_t *child, storage_object_t *new_child)
{
	int rc;
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	rc = md_replace_child(region, child, new_child);
	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_add_sectors_to_kill_list(storage_object_t *region, lsn_t lsn, sector_count_t count)
{
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	LOG_EXIT_INT(0);
	return 0;
}

static int raid10_commit_changes(storage_object_t *region, uint phase)
{
	md_volume_t * vol;
	int rc = 0;
	int saved_rc;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	if (!region)
	{
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	if (region->plugin != raid10_plugin)
	{
		LOG_ERROR("Region %s does not belong to MD.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}
	if (! (region->flags & SOFLAG_DIRTY))
	{
		LOG_WARNING("Region %s is not dirty - not committing.\n", region->name);
		LOG_EXIT_INT(0);
		return 0;
	}

	vol = (md_volume_t *)region->private_data;
	if (!vol)
	{
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

//XXX
	switch (phase) {
	case SETUP:
		rc = process_setup_funcs(region);
		break;
	case FIRST_METADATA_WRITE:
		if (vol->flags & MD_DIRTY)
		{
			rc = md_write_sbs_to_disk(vol);
			vol->flags &= ~MD_NEW_REGION;
		}
		break;
	case SECOND_METADATA_WRITE:
		break;
	case POST_ACTIVATE:
		rc = process_md_ioctl_pkgs(region);
		free_ioctl_pkgs(vol);

		/* Despite rc, we will rediscover the MD array, save the return code */
		saved_rc = rc;
		rc = raid10_rediscover_region(region, TRUE);
		if (!rc)
		{
			/*
			 * The region was rediscovered, volume pointer is now invalid.
			 * The correct volume pointer is region->private_data.
			 */
			vol = (md_volume_t *)region->private_data;
			region->flags &= ~SOFLAG_DIRTY;
		}
		else
		{
			/* Rediscovery failed, display any queued corrupt messages. */
			md_display_corrupt_messages(RAID10);
		}

		if (saved_rc != 0)
			rc = saved_rc;
		break;
	default	:
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_can_activate_region(storage_object_t *region)
{
	md_volume_t   * volume = (md_volume_t *)region->private_data;
	my_plugin = raid10_plugin;
	LOG_ENTRY();

	if (volume->flags & MD_CORRUPT) {
		LOG_WARNING("MD region %s is corrupt.\n", volume->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_EXIT_INT(0);
	return 0;
}

static int raid10_activate_region(storage_object_t *region)
{
	int rc;
	md_volume_t *vol;
	my_plugin = raid10_plugin;
	LOG_ENTRY();

	if (!region || !(vol = (md_volume_t *)region->private_data)) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	rc = md_activate_region (region);

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_can_deactivate_region(storage_object_t *UNUSED(region))
{
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	LOG_EXIT_INT(0);
	return 0;
}

static int raid10_deactivate_region(storage_object_t *region)
{
	int rc=0;
	my_plugin = raid10_plugin;
	LOG_ENTRY();

	rc = md_deactivate_region (region);

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_get_option_count(task_context_t *task)
{
	int count = 0;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	switch (task->action) {
	case EVMS_Task_Create:
		count = RAID10_CREATE_OPT_COUNT;
		break;
	default:
		count = 0;
		break;
	}

	LOG_EXIT_INT(count);
	return count;
}

static int raid10_init_task(task_context_t *context)
{
	int rc = 0;
	list_anchor_t tmp_list;
	md_volume_t *vol;
	md_member_t *member;
	list_element_t iter;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	// Parameter check.
	if (!context) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	switch (context->action) {

	case EVMS_Task_Create:

		context->option_descriptors->count = RAID10_CREATE_OPT_COUNT;

		// Version 1 Superblock Option
		if (md_can_create_sb_1() == TRUE) {
			context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].flags = 0;
			context->min_selected_objects = 2;
			context->max_selected_objects = MD_SB_1_DISKS;
		} else {
			context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].flags = EVMS_OPTION_FLAGS_INACTIVE;
			context->min_selected_objects = 2;
			context->max_selected_objects = MD_SB_DISKS;
		}
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].constraint.list = NULL;
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].constraint_type = EVMS_Collection_None;
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].help = NULL;
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].name =
			EngFncs->engine_strdup( RAID10_CREATE_OPT_SB1_NAME );
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].tip =
			EngFncs->engine_strdup( _("Choose Yes if you want to create MD version 1 super block.") );
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].title = EngFncs->engine_strdup( _("Version 1 Super Block") );
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].type = EVMS_Type_Boolean;
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].unit = EVMS_Unit_None;
		context->option_descriptors->option[RAID10_CREATE_OPT_SB1_INDEX].value.b = FALSE;

		// Chunk Size Option
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].flags = 0;
		SET_POWER2_LIST(context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].constraint.list, MD_MIN_CHUNK_SIZE, MD_MAX_CHUNK_SIZE);
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].constraint_type = EVMS_Collection_List;
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].help = NULL;
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].name = EngFncs->engine_strdup(RAID10_CREATE_OPT_CHUNK_SIZE_NAME);
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].tip = EngFncs->engine_strdup(_("Size of IO to each member of the array (also refered to as stripe size)." ));
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].title = EngFncs->engine_strdup(_("Chunk Size:" ));
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].type = EVMS_Type_Unsigned_Int32;
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].unit = EVMS_Unit_Kilobytes;
		context->option_descriptors->option[RAID10_CREATE_OPT_CHUNK_SIZE_INDEX].value.ui32 = MD_DEFAULT_CHUNK_SIZE;

		// Layout Near Option
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].constraint.list = NULL;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].constraint.range = EngFncs->engine_alloc( sizeof(value_range_t) );
                if (context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].constraint.range==NULL)
		{
                        LOG_EXIT_INT(ENOMEM);
                        return ENOMEM;
                }
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].constraint_type = EVMS_Collection_Range;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].constraint.range->min.ui8 = 0x02;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].constraint.range->max.ui8 = 0xff;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].constraint.range->increment.ui8 = 1;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].flags = 0;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].help = NULL;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].name = EngFncs->engine_strdup(RAID10_CREATE_OPT_LAYOUT_NEAR_NAME);
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].tip = EngFncs->engine_strdup( _("Multiple copies of one data block are at similar offsets in different devices.") );
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].title = EngFncs->engine_strdup( _("Layout (near):") );
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].type = EVMS_Type_Unsigned_Int8;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].unit = EVMS_Unit_None;
                context->option_descriptors->option[RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX].value.ui8 = RAID10_DEFAULT_LAYOUT_NEAR;

		// get a list of all valid input disks, segments, and regions.
		EngFncs->get_object_list(DISK | SEGMENT | REGION,
					 DATA_TYPE,
					 NULL,
					 NULL,
					 VALID_INPUT_OBJECT,
					 &tmp_list);

		// move these items to the acceptable objects list.
		md_transfer_list(tmp_list, context->acceptable_objects);
		EngFncs->destroy_list(tmp_list);

		break;

	default:
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_set_option(task_context_t *context, u_int32_t index, value_t *value, task_effect_t *effect)
{
	int rc = 0;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	// Parameter check.
	if (!context || !value || !effect) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	switch (context->action) {

	case EVMS_Task_Create:
		switch (index) {
		case RAID10_CREATE_OPT_SB1_INDEX:
			context->option_descriptors->option[index].value.b = value->b;
			if (value->b == TRUE) {
				context->max_selected_objects = MD_SB_1_DISKS;
			} else {
				context->max_selected_objects = MD_SB_DISKS;
			}
			break;

		case RAID10_CREATE_OPT_CHUNK_SIZE_INDEX:
			// Option is in kilobytes
			context->option_descriptors->option[index].value.ui32 = value->ui32;
			break;

		case RAID10_CREATE_OPT_LAYOUT_NEAR_INDEX:
			context->option_descriptors->option[index].value.ui8 = value->ui8;
			break;

		default:
			break;
                }

	default:
		break;
	}
	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_set_objects(task_context_t *context, list_anchor_t declined_objects, task_effect_t *effect)
{
	int rc = 0;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	// Parameter check.
	if (!context || !declined_objects || !effect) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	switch (context->action) {

	case EVMS_Task_Create:
		// Assume front end has selected valid objects from the
		// acceptable_objects list.  Could add safety checking.
		break;

	default:
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_get_info(storage_object_t *region, char *name, extended_info_array_t **info_array)
{
	md_volume_t * volume = NULL;
	int           rc= 0;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	// Parameter check
	if (!info_array) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	// Make sure this is an MD region
	if (region->plugin != raid10_plugin) {
		LOG_ERROR("Region %s is not owned by MD RAID1\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	volume = region->private_data;

	rc = md_get_info(volume, name, info_array);

	LOG_EXIT_INT(rc);
	return rc;
}

static int raid10_get_plugin_info(char *name, extended_info_array_t **info_array)
{
	extended_info_array_t   * info = NULL;
	char                    buffer[50] = {0};
	int                     i = 0;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	// Parameter check
	if (!info_array) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	if (!name) {
		// Get memory for the info array
		if (!(info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*6))) {
			LOG_ERROR("Error allocating memory for info array\n");
			LOG_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		// Short Name
		info->info[i].name = EngFncs->engine_strdup("ShortName");
		info->info[i].title = EngFncs->engine_strdup(_("Short Name"));
		info->info[i].desc = EngFncs->engine_strdup(_("A short name given to this plug-in"));
		info->info[i].type = EVMS_Type_String;
		info->info[i].value.s = EngFncs->engine_strdup(raid10_plugin->short_name);
		i++;

		// Long Name
		info->info[i].name = EngFncs->engine_strdup("LongName");
		info->info[i].title = EngFncs->engine_strdup(_("Long Name"));
		info->info[i].desc = EngFncs->engine_strdup(_("A longer, more descriptive name for this plug-in"));
		info->info[i].type = EVMS_Type_String;
		info->info[i].value.s = EngFncs->engine_strdup(raid10_plugin->long_name);
		i++;

		// Plugin Type
		info->info[i].name = EngFncs->engine_strdup("Type");
		info->info[i].title = EngFncs->engine_strdup(_("Plug-in Type"));
		info->info[i].desc = EngFncs->engine_strdup(_("There are various types of plug-ins, each responsible for some kind of storage object or logical volume."));
		info->info[i].type = EVMS_Type_String;
		info->info[i].value.s = EngFncs->engine_strdup(_("Region Manager"));
		i++;

		// Plugin Version
		info->info[i].name = EngFncs->engine_strdup("Version");
		info->info[i].title = EngFncs->engine_strdup(_("Plug-in Version"));
		info->info[i].desc = EngFncs->engine_strdup(_("This is the version number of the plug-in."));
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);
		info->info[i].value.s = EngFncs->engine_strdup(buffer);
		i++;

		// Required Engine Services Version
		info->info[i].name = EngFncs->engine_strdup("Required_Engine_Version");
		info->info[i].title = EngFncs->engine_strdup(_("Required Engine Services Version"));
		info->info[i].desc = EngFncs->engine_strdup(_("This is the version of the Engine services that this plug-in requires.  "
							      "It will not run on older versions of the Engine services."));
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", raid10_plugin->required_engine_api_version.major, raid10_plugin->required_engine_api_version.minor, raid10_plugin->required_engine_api_version.patchlevel);
		info->info[i].value.s = EngFncs->engine_strdup(buffer);
		i++;

		// Required Plug-in API Version
		info->info[i].name = EngFncs->engine_strdup("Required_Plugin_Version");
		info->info[i].title = EngFncs->engine_strdup(_("Required Plug-in API Version"));
		info->info[i].desc = EngFncs->engine_strdup(_("This is the version of the Engine plug-in API that this plug-in requires.  "
							      "It will not run on older versions of the Engine plug-in API."));
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", raid10_plugin->required_plugin_api_version.plugin.major, raid10_plugin->required_plugin_api_version.plugin.minor, raid10_plugin->required_plugin_api_version.plugin.patchlevel);
		info->info[i].value.s = EngFncs->engine_strdup(buffer);
		i++;
	} else {
		LOG_ERROR("No support for extra plugin information about \"%s\"\n", name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	info->count = i;
	*info_array = info;
	LOG_EXIT_INT(0);
	return 0;
}

static int raid10_read(storage_object_t *region, lsn_t lsn, sector_count_t count, void *buffer)
{
	md_volume_t * vol;
	md_member_t *member;
	list_element_t iter;
	int rc = 0;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	if (!region || !buffer) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	vol = (md_volume_t *)region->private_data;
	if (!vol) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	if (vol->flags & MD_CORRUPT) {
		memset(buffer, 0x0, count * EVMS_VSECTOR_SIZE);
		LOG_ERROR("MD Object %s is corrupt, returning zero filled buffer.\n",
			  vol->name);
		LOG_EXIT_INT(0);
		return 0;
	}

	if ((lsn + count) > region->size) {
		LOG_ERROR("Attempt to read past end of region %s sector=%"PRIu64"\n",
			  vol->name, lsn+count);
	}

	rc = md_region_rw(region, lsn, count, buffer, 0);
//XXX
	LOG_EXIT_INT(0);
	return 0;
}

static int raid10_write(storage_object_t *region, lsn_t lsn, sector_count_t count, void *buffer)
{
	int                rc = 0;
	md_volume_t      * volume = (md_volume_t *)region->private_data;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	// Parameter check.
	if (!buffer) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	if (volume->flags & MD_CORRUPT) {
		MESSAGE(_("MD Object %s is corrupt.  Writing data is not allowed.\n"),volume->name);
		LOG_EXIT_INT(EIO);
		return EIO;
	}

	if ((lsn + count) > region->size) {
		LOG_ERROR("Attempt to write past end of region %s sector=%"PRIu64"\n",volume->name,lsn+count);
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	rc = md_region_rw(region, lsn, count, buffer, 1);
//XXX
	LOG_EXIT_INT(0);
	return 0;
}

static int raid10_get_plugin_functions(storage_object_t *UNUSED(region), function_info_array_t **UNUSED(functions))
{
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}

static int raid10_plugin_function(storage_object_t *UNUSED(object), task_action_t UNUSED(action), list_anchor_t UNUSED(objects), option_array_t *UNUSED(options))
{
	my_plugin = raid10_plugin;
	LOG_ENTRY();
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}

static void free_region(storage_object_t *region)
{
	md_volume_t * volume = (md_volume_t *)region->private_data;
	LOG_ENTRY();
	md_free_volume(volume);
	LOG_EXIT_VOID();
}

int raid10_backup_metadata(storage_object_t *region)
{
	md_volume_t *volume;
	int rc=0;

	my_plugin = linear_plugin;
	LOG_ENTRY();


	volume = region->private_data;
	if (volume->flags & MD_CORRUPT) {
		rc = ENOSYS;
		goto out;
	}
	volume->commit_flag |= MD_COMMIT_BACKUP_METADATA;
	volume->commit_flag |= MD_COMMIT_DONT_CHECK_ACTIVE;
	volume->flags |= MD_DIRTY;
	rc = md_write_sbs_to_disk(volume);
	volume->commit_flag &= ~MD_COMMIT_BACKUP_METADATA;
	volume->commit_flag &= ~MD_COMMIT_DONT_CHECK_ACTIVE;

out:
	LOG_EXIT_INT(rc);
	return rc;
}

static void raid10_plugin_cleanup()
{
	list_anchor_t regions_list;
	list_element_t li;
	storage_object_t *region;

	my_plugin = raid10_plugin;
	LOG_ENTRY();

	EngFncs->get_object_list(REGION, DATA_TYPE, raid10_plugin, NULL, 0, &regions_list);

	LIST_FOR_EACH(regions_list, li, region)
	{
		free_region(region);
	}

	EngFncs->destroy_list(regions_list);

	LOG_EXIT_VOID();
}

static plugin_functions_t raid10_functions = {
	setup_evms_plugin              : raid10_setup_evms_plugin,
	cleanup_evms_plugin            : raid10_plugin_cleanup,
	can_delete                     : raid10_can_delete,
	can_expand                     : raid10_can_expand,
	can_shrink                     : raid10_can_shrink,
	can_replace_child              : raid10_can_replace_child,
	discover                       : raid10_discover,
	create                         : raid10_create,
	delete                         : raid10_delete,
	discard                        : raid10_discard,
	expand                         : raid10_expand,
	shrink                         : raid10_shrink,
	replace_child                  : raid10_replace_child,
	add_sectors_to_kill_list       : raid10_add_sectors_to_kill_list,
	commit_changes                 : raid10_commit_changes,
	can_activate                   : raid10_can_activate_region,
	activate                       : raid10_activate_region,
	can_deactivate                 : raid10_can_deactivate_region,
	deactivate                     : raid10_deactivate_region,
	get_option_count               : raid10_get_option_count,
	init_task                      : raid10_init_task,
	set_option                     : raid10_set_option,
	set_objects                    : raid10_set_objects,
	get_info                       : raid10_get_info,
	get_plugin_info                : raid10_get_plugin_info,
	read                           : raid10_read,
	write                          : raid10_write,
	get_plugin_functions           : raid10_get_plugin_functions,
	plugin_function                : raid10_plugin_function,
	backup_metadata                : raid10_backup_metadata
};

plugin_record_t raid10_plugin_record = {
	id:                     EVMS_MD_RAID10_PLUGIN_ID,

	version:                {major:      MAJOR_VERSION,
				 minor:      MINOR_VERSION,
				 patchlevel: PATCH_LEVEL},

	required_engine_api_version: {major:      15,
				      minor:      0,
				      patchlevel: 0},
	required_plugin_api_version: {plugin: {major:      13,
					       minor:      0,
					       patchlevel: 0} },

	short_name:             EVMS_MD_RAID10_PLUGIN_SHORT_NAME,
	long_name:              EVMS_MD_RAID10_PLUGIN_LONG_NAME,
	oem_name:               EVMS_ALT_LINUX_OEM_NAME,

	functions:              {plugin: &raid10_functions},

	container_functions:    NULL
};

